/*
 *  Wine Fridge firmware
 * 
 *  Copyright (c) 2020 dUkk
*/
#include <Wire.h>

//#define DEBUG 1

#define TEMPOFFSET -2

#define FRIDGE_OFF 0
#define FRIDGE_ON 1
#define FRIDGE_SETTEMP 2


//DISPLAY stuff (Digit write selection connected to PD3 PD4 PD5, data PD6, clock PD7)
#define DISP_BACKLIGHT_PIN 2
#define DISP_DIGIT0_PIN 3
#define DISP_DIGIT1_PIN 4
#define DISP_DIGIT2_PIN 5
#define DISP_DATA_PIN 6
#define DISP_CLOCK_PIN 7

//NTC THERMISTOR stuff  (connected to ADC0 = pin A0)
#define THERMISTOR_BETA 3950 //thermistor Beta coeff
#define SERIAL_R 10000 //divider resistance (10k)
#define THERMISTOR_R 10000 //thermistor nominal resistance at 25 C (10k)
#define NOMINAL_T 25 //nominal temperature at which thermistor has resistance of (10k)

//peltier cooler connected to PB0, FAN internal PB1, light PB2
#define PELTIER_PIN 8
#define FAN1_PIN 9
#define LIGHT_PIN 10

//filter 
#define FILTERSIZE 30
int8_t FilterTaps[FILTERSIZE], FilterIndex = 0, FilterFill = 0;
int16_t FilterSum = 0;


//defaults
int8_t chamberReqTemp=20, DisplayValue=0, WorkingMode=FRIDGE_OFF;
uint8_t ModeChangeCounter=0;
bool LightState=false;
unsigned long ts, WorkCycleMS = 1000;
byte Digval[11] = {B11111100,  //0
                   B00001100,  //1
                   B11011010,  //2
                   B10011110,  //3
                   B00101110,  //4
                   B10110110,  //5
                   B11110110,  //6
                   B00111100,  //7
                   B11111110,  //8
                   B10111110,  //9
                   B00000000}; //zero all
void DigitWrite(byte value);
int8_t GetChamberTemp();
int8_t GetPressedButton();
void SetLightState(bool NewState);
void SetFan1State(bool NewState);
void SetPeltierState(bool NewState);
void SetDisplayBacklightState(bool NewState);
void AT24C_WriteByte(byte dev, byte Address, byte data);
byte AT24C_ReadByte(int dev, byte Address);
int16_t ADC_read(uint8_t pin);

void setup()
{
#ifdef DEBUG
  Serial.begin(9600);
  delay(10);
#endif

  Wire.begin();
  
  //TODO DDRD |= (1<<PDx) | ...;
  pinMode(DISP_BACKLIGHT_PIN, OUTPUT);
  pinMode(DISP_DIGIT0_PIN, OUTPUT);
  pinMode(DISP_DIGIT1_PIN, OUTPUT);
  pinMode(DISP_DIGIT2_PIN, OUTPUT);
  pinMode(DISP_CLOCK_PIN, OUTPUT);
  pinMode(DISP_DATA_PIN, OUTPUT);
  pinMode(LIGHT_PIN, OUTPUT);
  pinMode(PELTIER_PIN, OUTPUT);
  pinMode(FAN1_PIN, OUTPUT);
  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);

  //enable ADC and set prescaler
  // ADPS2:ADPS1:ADPS0
  //001 СК/2
  //010 СК/4
  //011 СК/8
  //100 СК/16
  //101 СК/32
  //110 СК/64
  //111 СК/128
  ADCSRA |= (1 << ADPS2);                     //set 1
  ADCSRA &= ~ ((1 << ADPS1) | (1 << ADPS0));  //set 0
  ADCSRA |= (1 << ADEN);                      //enable ADC
  
  //display prepare
  PORTD |= _BV(PD3); // HIGH  //dig0
  PORTD |= _BV(PD4); // HIGH  //dig1
  PORTD |= _BV(PD5); // HIGH  //dig2
  PORTD |= _BV(PD6); // HIGH  //data
  PORTD &= ~ _BV(PD7); // LOW //clock

#ifdef DEBUG
  for (int8_t i = 0; i < 2; i++)
  {
    Serial.print(AT24C_ReadByte(0x50, i), HEX);
    Serial.print(" ");
  }
  Serial.println();
#endif
 
  int8_t tmp;
  //try load our power state
  tmp = AT24C_ReadByte(0x50, 0);
  if(tmp == 1) 
  {
    WorkingMode = FRIDGE_ON;
    //turn on backlight
    SetDisplayBacklightState(true);
    //fill filter with initial temp readings
    for(int8_t i=0;i<FILTERSIZE;i++) DisplayValue = GetChamberTemp();
  }
  else {
    SetDisplayBacklightState(false);
  }
  
  //try load our saved temperature
  tmp = AT24C_ReadByte(0x50, 1);
  if(tmp >= 4 && tmp <= 25) chamberReqTemp = tmp;

  SetLightState(false);
  SetFan1State(false);
  SetPeltierState(false);
  
  ts = millis();
}

void loop() 
{

  if(WorkingMode == FRIDGE_OFF)
  {
    //=========DIGIT0=========
 
    //unknown clock cycle
    for(int b=0;b<16;b++)
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits
    DigitWrite(Digval[10]);

    //place digit0 line low
    PORTD &= ~ _BV(PD3); // LOW

    delay(2);// wait 4ms  
    
    
    //=========DIGIT1=========
    PORTD |= _BV(PD3); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits
    DigitWrite(Digval[10]);

    PORTD &= ~ _BV(PD4); // LOW

    delay(2);// wait 4ms  
    

    //=========DIGIT2=========
    PORTD |= _BV(PD4); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits
    DigitWrite(Digval[10]);

    PORTD &= ~ _BV(PD5); // LOW

    delay(2);// wait 4ms  
    
    //ending sequence
    PORTD |= _BV(PD5); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }

    delay(2);//wait 4ms    
  }
  //in other modes display should be ON and display some values
  else {
    //=========DIGIT0=========
 
    //unknown clock cycle
    for(int b=0;b<16;b++)
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits
    DigitWrite(Digval[(DisplayValue % 10)]);

    //place digit0 line low
    PORTD &= ~ _BV(PD3); // LOW

    delay(2);// wait 4ms  
    
    
    //=========DIGIT1=========
    PORTD |= _BV(PD3); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits
    DigitWrite(Digval[((DisplayValue / 10) % 10)]);

    PORTD &= ~ _BV(PD4); // LOW

    delay(2);// wait 4ms  
    

    //=========DIGIT2=========
    PORTD |= _BV(PD4); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }
    delayMicroseconds(15); //wait 30us

    //write digit bits (encoded value for Celsius)
    //do blinking: odd number is C , even number is empty
    if((ModeChangeCounter & 1) == 0) DigitWrite(B11110001);
    else DigitWrite(Digval[10]);
        

    PORTD &= ~ _BV(PD5); // LOW

    delay(2);// wait 4ms  
    
    //ending sequence
    PORTD |= _BV(PD5); // HIGH
  
    PORTD |= _BV(PD6); // HIGH
  
    //unknown clock cycle
    for(int b=0;b<16;b++) 
    {
      PORTD ^= _BV(PD7); //toggle
      delayMicroseconds(3); //wait 3us
    }

    delay(2);//wait 4ms
  }
  
  //do we need to perform any actions?
  unsigned long curts = millis();
  if((curts - ts) >= WorkCycleMS)
  {
    ts = curts;

    int16_t buttonread = ADC_read(A1);
  #ifdef DEBUG    
        Serial.println(buttonread);
  #endif              
    
    //in this mode we only read power button
    if(WorkingMode == FRIDGE_OFF)
    {   
      if(buttonread >= 800 && buttonread <= 840)
      {
        //start turning fridge ON
        WorkingMode = FRIDGE_ON;
        WorkCycleMS = 1000;
        //turn on backlight
        SetDisplayBacklightState(true);
        
        //save working mode on
        if(AT24C_ReadByte(0x50, 0) != 1)
        {
  #ifdef DEBUG    
            Serial.println("SAVE PWRON");
  #endif
            AT24C_WriteByte(0x50, 0, 1);
        }

        //fill filter with initial temp readings
        for(int8_t i=0;i<FILTERSIZE;i++) DisplayValue = GetChamberTemp();

  #ifdef DEBUG    
        Serial.println("FRIDGE_ON");
  #endif        
      }
    }
    //any working mode of fridge here
    else
    {
      //POWER
      if(buttonread >= 800 && buttonread <= 840)
      {
        WorkingMode = FRIDGE_OFF;
        WorkCycleMS = 1500;
        //turn off all
        SetDisplayBacklightState(false);
        SetFan1State(false);
        SetLightState(false);
        SetPeltierState(false);
        LightState = false;
  #ifdef DEBUG    
        Serial.println("FRIDGE_OFF");
  #endif
        //save working mode off
        if(AT24C_ReadByte(0x50, 0) != 0)
        {
  #ifdef DEBUG    
            Serial.println("SAVE PWROFF");
  #endif
            AT24C_WriteByte(0x50, 0, 0);
        }
      }
      //LIGHT
      else if(buttonread >= 630 && buttonread <= 655)
      {
        LightState = !LightState;
  #ifdef DEBUG    
        Serial.println("LIGHT_TOGGLE");
  #endif              
      }
      //TEMP-
      else if(buttonread >= 410 && buttonread <= 460)
      {
        WorkCycleMS = 200;
        ModeChangeCounter = 0;
        WorkingMode = FRIDGE_SETTEMP;
        if(chamberReqTemp > 4) chamberReqTemp--;
  #ifdef DEBUG    
        Serial.println("TEMP+");
  #endif              
      }
      //TEMP+
      else if(buttonread >= 220 && buttonread <= 275)
      {
        WorkCycleMS = 200;
        ModeChangeCounter = 0;
        WorkingMode == FRIDGE_SETTEMP;
        if(chamberReqTemp < 25) chamberReqTemp++;
  #ifdef DEBUG    
        Serial.println("TEMP-");
  #endif              
      }

      if(WorkingMode == FRIDGE_SETTEMP)
      {
        DisplayValue = chamberReqTemp;
        if(ModeChangeCounter < 25) ModeChangeCounter++;
        else
        {
  #ifdef DEBUG    
        Serial.println("ENDING_SETTEMP");
  #endif
          ModeChangeCounter=0;
          WorkCycleMS = 1000;
          WorkingMode = FRIDGE_ON;

          //save temperature if its changed
          if(AT24C_ReadByte(0x50, 1) != chamberReqTemp)
          {
  #ifdef DEBUG    
            Serial.println("SAVE TEMP");
  #endif
            AT24C_WriteByte(0x50, 1, chamberReqTemp);
          }
        }
      }
      else if(WorkingMode == FRIDGE_ON)
      {  
        //obtain current temperature
        DisplayValue = GetChamberTemp();
        //do peltier logic here. 
        //NOTICE: we don't need hysteresis here (+-2 degree) because our N-tap filter has slow response already
        if(DisplayValue > chamberReqTemp)
        {
          SetPeltierState(true);
  #ifdef DEBUG    
        Serial.println("SetPeltierState true");
  #endif             
        }
        else if(DisplayValue < chamberReqTemp)
        {
          SetPeltierState(false);
        }
        
  #ifdef DEBUG    
        Serial.println(DisplayValue);
  #endif              
      }

      //door is OPENED
      if(digitalRead(A2) == HIGH)
      {
        SetLightState(true);
        SetFan1State(false);
  #ifdef DEBUG    
        Serial.println("DOOR_OPEN");
  #endif     
      }
      //door is CLOSED
      else 
      {
        //restore last state of light and turn on fan
        SetLightState(LightState);
        SetFan1State(true);
      }
      
    }
  
  }
 
}


void DigitWrite(byte value)
{
  if( ( value & (1<< 0)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 1)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 2)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 3)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 4)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 5)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 6)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
  delayMicroseconds(3); //wait 3us
  if( ( value & (1<< 7)) ) PORTD &= ~ _BV(PD6); // LOW
  PORTD |= _BV(PD7); // HIGH
  delayMicroseconds(3); //wait 3us
  //PORTD |= _BV(PD6); // HIGH
  PORTD &= ~ _BV(PD7); // LOW
}

int8_t GetChamberTemp()
{
  //do reading raw ADC value and process it with classic NTC-thermistor formula
  float steinhart = 1023.0 / ADC_read(A0) - 1;
  steinhart = SERIAL_R / steinhart;
  steinhart = steinhart / THERMISTOR_R; // (R/Ro)
  steinhart = log(steinhart); // ln(R/Ro)
  steinhart /= THERMISTOR_BETA; // 1/B * ln(R/Ro)
  steinhart += 1.0 / (NOMINAL_T + 273.15); // + (1/To)
  steinhart = 1.0 / steinhart; // Invert
  steinhart -= 273.15; //F to Celsius

  //now do moving average filter on returned temperature
  if (FilterFill)
  {
    FilterSum -= FilterTaps[FilterIndex];
  }  
  FilterTaps[FilterIndex] = lrintf(steinhart);
  FilterSum += FilterTaps[FilterIndex];
  FilterIndex++;
  if (FilterIndex >= FILTERSIZE)
  {
    FilterIndex = 0;
  }
  if (FilterFill < FILTERSIZE)
  {
    FilterFill++;
  }

  //return average value in our array  plus do temperature offset corrections
  return (FilterSum / FilterFill) + TEMPOFFSET;
}

void SetLightState(bool NewState)
{
  if(NewState) PORTB |= _BV(PB2); // HIGH
  else PORTB &= ~ _BV(PB2); // LOW
}

void SetFan1State(bool NewState)
{
  if(NewState) PORTB |= _BV(PB1); // HIGH
  else PORTB &= ~ _BV(PB1); // LOW
}

void SetPeltierState(bool NewState)
{
  if(NewState) PORTB |= _BV(PB0); // HIGH
  else PORTB &= ~ _BV(PB0); // LOW
}

void SetDisplayBacklightState(bool NewState)
{
  if(NewState) PORTD |= _BV(PD2); // HIGH
  else PORTD &= ~ _BV(PD2); // LOW  
}

void AT24C_WriteByte(byte dev, byte Address, byte data)
{
  Wire.beginTransmission(dev);
  Wire.write(Address >> 8);
  Wire.write(Address & 0xFF);
  Wire.write(data);
  Wire.endTransmission();
  delay(5);
}

byte AT24C_ReadByte(int dev, byte Address) 
{
  byte rdata = 0xFF;
  Wire.beginTransmission(dev);
  Wire.write(Address >> 8);
  Wire.write(Address & 0xFF); 
  Wire.endTransmission();
  Wire.requestFrom(dev, 1);
  if (Wire.available()) rdata = Wire.read();
  return rdata;
}

int16_t ADC_read(uint8_t pin)
{
  //initial signal measurement and noise filtering (8 points average filter)
    uint8_t i, high,low;
    uint16_t ADC_var = 0;

    pin -= 14; //on mega328 ports A0-A5 has >=14 pin number
    ADMUX = (0 << REFS1)|(1 << REFS0) | (pin & 0x07);

    for(i=0; i<8; i++)
    {
        ADCSRA |= _BV(ADSC);       // start conversion
        while(ADCSRA & _BV(ADSC)); // wait for conversion done, ADSC flag active
        ADC_var += ADCW;           // accumulate result for later averaging
    }

    //ADC_var = ADC_var / 8;         // average the 8 samples
    ADC_var >>= 3;

    return ADC_var;
}
